/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include <stdio.h>


mfc_list_element_t list[8];

char ls_buffer[65536] __attribute__ ((aligned(128)));


/* SPE program that demonstrates a few of the common race conditions
 * detectable by SW race checker library. The race conditions can
 * be repaired by compiling this example with FIX_RACE_CONDITIONS
 * defined.
 */

int main(unsigned long long speid __attribute__ ((unused)),
	 unsigned long long ea_buffer,
	 unsigned long long envp  __attribute__ ((unused)))
{
  mfc_list_element_t *plist;
  char c;

  mfc_write_tag_mask(0xFFFFFFFF);

  /* Race Condition 1 - Ordering two DMAs in separate tag groups. 
   * A fenced or barriered DMA ensures ordering only for DMAs in 
   * the same tag group. In this case, two DMAs were expected to 
   * ordered, however, they were inadvertantly issued using different
   * tag IDs (0 and 1). The fix to this problem is to use the same
   * tag ID (0).
   */
  printf("RACE CONDITION 1\n");
  mfc_get(&ls_buffer[0], ea_buffer, 16384, 0, 0, 0);
#ifdef FIX_RACE_CONDITIONS
  mfc_putb(&ls_buffer[32], ea_buffer+32768, 16, 0, 0, 0);
#else
  mfc_putb(&ls_buffer[32], ea_buffer+32768, 16, 1, 0, 0);
#endif
  
  mfc_read_tag_status_all();
  

  /* Race Condition 2 - Incorrect use of fence. In this case, we are 
   * copying a 32K system memory buffer to another system memory buffer.
   * Since the transfer must be broken into two DMAs, the two transfers 
   * can not be ordered using a fence. Fencing the first PUT only ensures
   * that it can not be initiated before the GETs complete. However, the
   * following PUT can be. Instead, a barrier should have been used.
   */
  printf("RACE CONDITION 2\n");
  mfc_get(&ls_buffer[0], ea_buffer, 16384, 1, 0, 0);
  mfc_get(&ls_buffer[16384], ea_buffer+16384, 16384, 1, 0, 0);
#ifdef FIX_RACE_CONDITIONS
  mfc_putb(&ls_buffer[0], ea_buffer+32768, 16384, 1, 0, 0);
#else
  mfc_putf(&ls_buffer[0], ea_buffer+32768, 16384, 1, 0, 0);
#endif
  mfc_put(&ls_buffer[16384], ea_buffer+49152, 16384, 1, 0, 0);

  mfc_read_tag_status_all();


  /* Race Condition 3 - Writing to a DMA list while a DMA may still be
   * in flight. In this example, we initiate a DMA GET list and then
   * proceed to inadvertantly reuse the list array before ensuring that
   * its previous DMA completed. The fix for this problem is to use
   * a separate list array or wait for DMA completion prior to constructing
   * the second list.
   */
  printf("RACE CONDITION 3\n");
  plist = list;
#ifdef _RACE_CHECK_H_
  RACE_CHECK_ACCESS(plist, 4*sizeof(mfc_list_element_t), RACE_CHECK_WRITE);
#endif
  plist[0].eal = mfc_ea2l(ea_buffer+0);   plist[0].size = 16;
  plist[1].eal = mfc_ea2l(ea_buffer+32);  plist[1].size = 16;
  plist[2].eal = mfc_ea2l(ea_buffer+64);  plist[2].size = 16;
  plist[3].eal = mfc_ea2l(ea_buffer+96);  plist[3].size = 16;
  mfc_getl(&ls_buffer[0], ea_buffer, plist, 4*sizeof(mfc_list_element_t), 1, 0, 0);

#ifdef FIX_RACE_CONDITIONS
  plist = &list[4];
#endif

#ifdef _RACE_CHECK_H_
  RACE_CHECK_ACCESS(plist, 4*sizeof(mfc_list_element_t), RACE_CHECK_WRITE);
#endif
  plist[0].eal = mfc_ea2l(ea_buffer+16);  plist[0].size = 16;
  plist[1].eal = mfc_ea2l(ea_buffer+48);  plist[1].size = 16;
  plist[2].eal = mfc_ea2l(ea_buffer+80);  plist[2].size = 16;
  plist[3].eal = mfc_ea2l(ea_buffer+112); plist[3].size = 16;
  mfc_getl(&ls_buffer[64], ea_buffer, plist, 4*sizeof(mfc_list_element_t), 1, 0, 0);

  mfc_read_tag_status_all();


  /* Race Condition 4 - Writing to a buffer that is being copied from local 
   * storage to system memory. In this case, the buffer is written before we 
   * tested that the DMA completed. The fix is to wait for the DMA to complete
   * before writing to the buffer.
   */ 
  printf("RACE CONDITION 4\n");
  mfc_put(&ls_buffer[0], ea_buffer, 16384, 1, 0, 0);

#ifdef FIX_RACE_CONDITIONS
  mfc_read_tag_status_all();
#endif

#ifdef _RACE_CHECK_H_
  RACE_CHECK_ACCESS(&ls_buffer[16383], 1, RACE_CHECK_WRITE);
#endif
  ls_buffer[16383] = 1;

#ifndef FIX_RACE_CONDITIONS
  mfc_read_tag_status_all();
#endif

  /* Race Condition 5 - Reading a buffer that is being copied from system memory
   * to local storage. In this case, the buffer is read before we tested
   * that the DMA completed. The fix is to wait for the DMA to complete before 
   * writing to the buffer.
   */ 
  printf("RACE CONDITION 5\n");
  mfc_get(&ls_buffer[0], ea_buffer, 16384, 1, 0, 0);

#ifdef FIX_RACE_CONDITIONS
  mfc_read_tag_status_all();
#endif

#ifdef _RACE_CHECK_H_
  RACE_CHECK_ACCESS(&ls_buffer[16383], 1, RACE_CHECK_WRITE);
#endif
  c = *((volatile char *)&ls_buffer[16383]);

#ifndef FIX_RACE_CONDITIONS
  mfc_read_tag_status_all();
#endif

  return 0;
}
